# -*- coding: utf-8 -*-

"""Top-level package for SQL in F(unctions)."""

__author__ = """Digitalmensch"""
__email__ = 'contact@digitalmensch.ch'
__version__ = '0.2.5'

from .sqlfunc import sqldb
from .sqlf import sql
from .sqlf import scalar_udf



import apsw
import inspect
import functools
import contextlib
import re

################################################################################
#
################################################################################

class sqldb(object):

    def __init__(self, sql=None, database=':memory:'):
        self.__db = apsw.Connection(database)
        if sql is not None:
            with contextlib.closing(self.__db.cursor()) as cur:
                cur.execute(sql)

    def sqludf(self, func):
        sig = inspect.signature(func)
        self.__db.createscalarfunction(func.__name__, func)

    def sqlfunc(self, _func=None, post=None, single=False, default=None):
        def _decorator(func):
            sql = func.__doc__
            assert len(sql) >= 8 # 'SELECT 1'
            signature = inspect.signature(func)
            params = tuple(signature.parameters.keys())

            @functools.wraps(func)
            def _function(*args, **kwargs):
                with contextlib.closing(self.__db.cursor()) as cur:
                    bound = signature.bind(*args, **kwargs)
                    bound.apply_defaults()
                    mapping = dict(zip(params, bound.args))

                    result = cur.execute(sql, mapping)

                    try:
                        description = result.getdescription()
                    except apsw.ExecutionCompleteError:
                        description = None
                        return default

                    if description:
                        keys = [
                            '_'.join(re.findall('\w[\w\d]+', attr[0]))
                            for attr in description
                        ]
                        result = map(lambda vals: dict(zip(keys, vals)), result)

                        if single:
                            for r in result:
                                return post(r) if post else r
                            else:
                                return default

                        else:
                            return list(map(post, result) if post else result)

                    else:
                        # self.__db.commit()
                        return self.__db.last_insert_rowid()

            return _function

        if _func is not None:
            return _decorator(_func)
        else:
            return _decorator



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `sqlf` package."""

import pytest
import sqlf

################################################################################
# Old API (sqlfunc)
################################################################################

db = sqlf.sqldb('''
    CREATE TABLE IF NOT EXISTS users (
        userid   INTEGER PRIMARY KEY,
        username TEXT    UNIQUE NOT NULL,
        bcrypt   BLOB    NOT NULL
    );
''', database=':memory:') # this is the default

@db.sqludf
def bcrypt_hash(password):
    # call to library here
    return b'$2b$12$.OjbRwRejxw92C89sA6JkOVrhmQzGsjoyCf1ofIN9hUNdHFufb3ty'

@db.sqludf
def bcrypt_verify(password, bcrypthash):
    # call to library here
    return 'password123' == password

@db.sqlfunc
def add_user(username, password):
    ''' INSERT OR IGNORE INTO users (username, bcrypt)
        VALUES (:username, bcrypt_hash(:password));
    '''

@db.sqlfunc(post=bool, single=True, default=False)
def login(username, password):
    ''' SELECT 1 FROM users
        WHERE username=:username
          AND bcrypt_verify(:password, bcrypt);
    '''

@db.sqlfunc
def list_users():
    ''' SELECT userid, username FROM users;
    '''

def test_scenario():
    '''Test script ported from sqlfunc'''

    add_user('root', 'password123')

    x1 = login('root', 'secret')
    print('x1', x1)
    assert False == x1

    x2 = login('root', 'password123')
    print('x2', x2)
    assert True == x2

    x3 = list_users()
    print('x3', x3)
    assert len(x3) == 1

################################################################################
# New API (sqlf)
################################################################################

def test_new():
    @sqlf.sql()
    def _setup():
        ''' create table tests (a, b, c, d); '''

    @sqlf.sql()
    def _insert(a, b=2, c=3):
        ''' insert into tests (a, b, c, d) values (:a, :b, :c, :c); '''

    @sqlf.sql()
    def _query():
        ''' select * from tests; '''

    assert 0 == len(list(_setup()))
    assert 0 == len(list(_query()))
    assert 0 == len(list(_insert(1, b=3)))
    assert 1 == len(list(_query()))

def test_scalar_udf():
    @sqlf.scalar_udf
    def one():
        return 1
    @sqlf.sql()
    def _select_one():
        ''' select one() as one; '''
    assert [{'one': 1}] == list(_select_one())



# -*- coding: utf-8 -*-

''' SQL in F(unctions)

    SQL in Python __doc__-strings as an alternative to ORMs
'''

import apsw
import atexit
import contextlib
import functools
import inspect
import re

################################################################################
# Globals / Initialisation
################################################################################
__connection = apsw.Connection(':memory:')
atexit.register(__connection.close)


################################################################################
# Public API
################################################################################

def sql(*, return_type=None):
    def _decorator(func):
        sql = func.__doc__
        signature = inspect.signature(func)
        params = tuple(signature.parameters.keys())
        @functools.wraps(func)
        def _wrapper(*args, **kwargs):
            with contextlib.closing(__connection.cursor()) as cursor:
                bound = signature.bind(*args, **kwargs)
                bound.apply_defaults()
                mapping = dict(zip(params, bound.args))
                cursor.execute(sql, mapping)
                try:
                    columns = list(map(colname, cursor.getdescription()))
                    for row in cursor:
                        tmp = dict(zip(columns, row))
                        yield return_type(tmp) if return_type else tmp
                except apsw.ExecutionCompleteError:
                    return
        return _wrapper
    return _decorator


def scalar_udf(func):
    __connection.createscalarfunction(func.__name__, func)
    return func

################################################################################
# Internals
################################################################################

def colname(desc):
    return '_'.join(re.findall('\w[\w\d]+', desc[0]))
